import unittest
from ...files.DictTTL_class import DictTTL
import time
from unittest.mock import patch
import math

class TestDictTTL(unittest.TestCase):
    def setUp(self):
        self.dict_ttl = DictTTL(10, a=1, b=2)

    @patch('time.sleep', return_value=None)
    def test_is_expired(self, patched_time_sleep):
        now = time.time()
        self.dict_ttl = DictTTL(10, a=1, b=2)
        self.assertFalse(self.dict_ttl.is_expired('a'))
        self.assertFalse(self.dict_ttl.is_expired('a',now=now))
        self.assertTrue(len(self.dict_ttl),2)
        time.sleep(10)
        self.assertTrue(self.dict_ttl.is_expired('a',now=now + 11))
        self.assertFalse(self.dict_ttl.is_expired('a', now=now + 9))
        self.assertTrue(len(self.dict_ttl), 1)

    @patch('time.sleep', return_value=None)
    def test__purge(self, patched_time_sleep):
        self.assertTrue(len(self.dict_ttl),2)
        time.sleep(11)
        self.dict_ttl._purge()
        self.assertTrue(len(self.dict_ttl),0)

    def test_get(self):
        self.assertEqual(self.dict_ttl.get('a')[1],1)
        self.assertEqual(self.dict_ttl.get('b')[1],2)

        self.assertNotEqual(self.dict_ttl.get('a')[1], 10)
        self.assertNotEqual(self.dict_ttl.get('b')[1], 20)


    def test_set_ttl(self):
        now = time.time()
        self.dict_ttl.set_ttl('a',10)
        self.assertEqual(math.floor(self.dict_ttl.get('a')[0]),math.floor(now + 10))
        self.assertNotEqual(math.floor(self.dict_ttl.get('a')[0]),math.floor(now + 11))

    def test_get_ttl(self):
        now = time.time()
        self.assertEqual(round(self.dict_ttl.get_ttl('a'),0),10)
        self.assertNotEqual(round(self.dict_ttl.get_ttl('a'),0),11)

    def test_expire_at(self):
        now = time.time()
        self.dict_ttl.expire_at('a',now + 10)
        self.assertEqual(math.ceil(self.dict_ttl.get_ttl('a')),10)
        self.assertNotEqual(math.ceil(self.dict_ttl.get_ttl('a')),11)

    def test__iter__(self):
        self.data = [('a',1),('b',2)]
        self.false_data = [('a',10),('b',30)]
        self.dict_data = list(self.dict_ttl.items())
        self.assertEqual(self.data,self.dict_data)
        self.assertNotEqual(self.data,self.false_data)

    def test__set_item(self):
        self.data = {'c':10,'d':20}
        self.data_items = [('c',10),('d',20)]
        self.false_data_items = [('c',30),('d',40)]
        self.dict_new_ttl = DictTTL(10,self.data)
        self.dict_data = list(self.dict_new_ttl.items())
        self.assertEqual(self.data_items, self.dict_data)
        self.assertNotEqual(self.false_data_items, self.dict_data)

    def test__delitem__(self):
        self.data = {'c':3,'d':4}
        self.assertIn('a', self.dict_ttl.keys())
        self.dict_ttl.__delitem__('a')
        self.assertNotIn('a',self.dict_ttl.keys())

    def test__len__(self):
        self.assertEqual(len(self.dict_ttl),2)
        self.assertNotEqual(len(self.dict_ttl),3)

    def test__getitem__(self):
        self.assertEqual(self.dict_ttl.__getitem__('a'),1)
        self.assertNotEqual(self.dict_ttl.__getitem__('b'),10)

    def test_keys(self):
        self.assertEqual(list(self.dict_ttl.keys()),['a','b'])
        self.assertNotEqual(list(self.dict_ttl.keys()), ['a', 'c'])

    def test_items(self):
        self.data = [('a',1),('b',2)]
        self.false_data = [('a',10),('b',20)]
        self.assertEqual(self.dict_ttl.items(),self.data)
        self.assertNotEqual(self.dict_ttl.items(), self.false_data)

    def test_ttl_items(self):
        now = time.time()
        self.data = [('a',(math.ceil(now) + 10,1)),('b',(math.ceil(now) + 10,2))]
        self.false_data = [('c',(math.ceil(now) + 20,10)),('d',(math.ceil(now) + 20,20))]

        [self.assertEqual(i[0],j[0]) for i,j in zip(self.dict_ttl.ttl_items(),self.data)]
        [self.assertEqual(math.ceil(i[1][0]),math.ceil(j[1][0])) for i, j in zip(self.dict_ttl.ttl_items(), self.data)]
        [self.assertEqual(i[1][1], j[1][1]) for i, j in zip(self.dict_ttl.ttl_items(), self.data)]

    def test_values(self):
        self.data = [1,2]
        self.false_data = [3,4]
        for k,v in zip(self.dict_ttl.values(),self.data):
            self.assertEqual(k[1],v)

        for k,v in zip(self.dict_ttl.values(),self.false_data):
            self.assertNotEqual(k[1],v)

    def test_sort_by_value(self):
        self.data = {5: 10, 4: 9, 3: 8 , 7:12 ,8: 14}
        self.dict_ttl = DictTTL(10, self.data).sort_by_value(reverse=False)
        for i,j in zip(self.dict_ttl.items(),sorted(list(self.data.items()))):
            self.assertEqual(i[1][0],j[1])
            self.assertNotEqual(i[1][0], 1)

    def test_values_without_ttl(self):
        self.assertEqual(self.dict_ttl.values_without_ttl(),[1,2])
        self.assertNotEqual(self.dict_ttl.values_without_ttl(),[1,2,3])

    def test_invert_dict_map(self):
        self.dict_ttl = self.dict_ttl.invert_dict_map()
        self.assertEqual(list(self.dict_ttl.keys()),[1,2])
        self.assertEqual(list(self.dict_ttl.values_without_ttl()),['a','b'])
        self.assertNotEqual(list(self.dict_ttl.keys()),['b','b'])
        self.assertNotEqual(list(self.dict_ttl.values_without_ttl()), [1, 2])

    def test_dict_union(self):
        self.data = {'c':3}
        self.assertEqual(DictTTL(time_to_live=10).dict_union(self.dict_ttl,self.data),{'a':1,'b':2,'c':3})
        self.assertNotEqual(DictTTL(time_to_live=10).dict_union(self.dict_ttl, self.data), {'a': 1, 'b': 2})

    def test_dict_intersection(self):
        self.data = {'b':2,'c': 3}
        self.assertEqual(DictTTL(time_to_live=10).dict_intersection(self.dict_ttl, self.data), {'b': 2})
        self.assertNotEqual(DictTTL(time_to_live=10).dict_intersection(self.dict_ttl, self.data), {'a':1,'b': 2})

if __name__ == '__main__':
    unittest.TestCase()



from setuptools import setup, find_packages

setup(name='dictttl',
      version='0.1.3',
      url='https://github.com/srivassid/DictTTL',
      license='MIT',
      author='Siddharth Srivastava',
      author_email='s.srivas@hotmail.com',
      description='Dictionary with TTL attached to the keys',
      packages=find_packages(),
      long_description=open('README.md').read(),
      zip_safe=False)


from threading import RLock
from collections import defaultdict
import time, math

class DefaultDictTTL(defaultdict):

    #initialize the dict with ttl and possibly values
    def __init__(self, time_to_live, *args, **kwargs):
        self.data = defaultdict(list)
        self._time_to_live = time_to_live
        self._lock = RLock()
        self.update(*args, **kwargs)

    def __repr__(self):
        self._purge()
        return '<DefaultDictTTL@%#08x; ttl=%r, Dict=%r;>' % (
            id(self), self._time_to_live, self.data)

    #check if key is expired by comparing value to now
    def is_expired(self, key, now=None):
        expire, _value = 0,0
        with self._lock:
            if now is None:
                now = time.time()
            try:
                expire, _value = self.data[key][0][0],self.data[key][0][1]

            except :
                if expire == 0 and _value == 0:
                    expire = time.time() + 10
            finally:
                if expire and expire < now:
                    return key

    #delete those keys which have been expired by checking is_expired method
    def _purge(self):
        _keys = list(self.data.keys())
        _remove = [key for key in _keys if self.is_expired(key)]
        for i in _remove:
            del self.data[i]

    #set ttl for a key making sure it is not being used by any other thread
    def set_ttl(self, key, ttl, now=None):
        if now is None:
            now = time.time()
        with self._lock:
            value = self[key][0][1]
            del self[key]
            self.data[key] =  [(now + ttl, value)]

    #get ttl for a key
    def get_ttl(self, key, now=None):
        if now is None:
            now = time.time()
        with self._lock:
            expire = self.data[key][0][0]
            return expire - now

    #set a manual expiration time for a key in epoch
    def expire_at(self, key, timestamp):
        with self._lock:
            value = self[key][0][1]
            del self[key]
            self.data[key] = [(timestamp, value)]

    #built in method that yields and iterator
    def __iter__(self):
        with self._lock:
            for key in self.data.keys():
                if not self.is_expired(key):
                    yield key

    #set a value for a key
    def __setitem__(self, key, value):
        with self._lock:
            if self._time_to_live is None:
                expire = None
            else:
                expire = time.time() + self._time_to_live
            # super().__setitem__(key, (expire, value))
            # self.data[key] = (expire, value)
            if len(self.data[key]) == 0:
                self.data[key].append((expire,[value]))
            else:
                self.data[key][0][1].append(value)

    def append_values(self, key, value):
        with self._lock:
            if self._time_to_live is None:
                expire = None
            else:
                expire = time.time() + self._time_to_live
            # super().__setitem__(key, (expire, value))
            # self.data[key] = (expire, value)
            if len(self.data[key]) == 0:
                self.data[key].append((expire,[value]))
            else:
                self.data[key][0][1].append(value)

    def append(self, key, value):
        print("for k,v in d:\n\tdict_ttl2.append_values(k,v)")

    #delete item from dict
    def __delitem__(self, key):
        with self._lock:
            del self.data[key]

    def delete_key_item(self,key, value):
        with self._lock:
            # if value in self.data[key][0][1]:
            try:
                self.data[key][0][1].remove(value)
            except Exception as e:
                print("Key not found " + str(e))

    #check length of dict after purging expired items
    def __len__(self):
        with self._lock:
            self._purge()
            return len(self.data.keys())

    #get the value for a particular key if it is not expired
    def __getitem__(self, key):
        with self._lock:
            if self.is_expired(key):
                try:
                    del self.data[key]
                except Exception as e:
                    print("Key not found " + str(KeyError))
        return self.data[key]

    #get keys after purging expired ones
    def keys(self):
        with self._lock:
            self._purge()
            return self.data.keys()

    #get key value pair without TTL
    def items(self):
        with self._lock:
            self._purge()
            return [(k,v[0][1]) for k,v in self.data.items()]

    #get key value pair in key, timestamp, order value
    def ttl_items(self):
        with self._lock:
            self._purge()
            return [(k,v) for k,v in self.data.items()]

    #get values with timestamp
    def values(self):
        with self._lock:
            self._purge()
            return [v[1] for v in self.data.items()]

    #get values withtout timestamp
    def values_without_ttl(self):
        with self._lock:
            self._purge()
            return [v[0][1] for k,v in self.data.items()]

    #get value for a key in O(1) time
    def get(self, key, default=None):
        try:
            return self.data[key]
        except KeyError:
            return default

    #union of items of two dicts
    def dict_union(self, dict1, dict2):
        with self._lock:
            self._purge()
            return {**dict1, **dict2}

    #intersection of items of two dicts
    def dict_intersection(self, dict1, dict2):
        with self._lock:
            self._purge()
            self.dict = defaultdict(list)
            self.common_keys = list([dict1.keys() & dict2.keys()][0])
            for i in self.common_keys:
                if dict1.get(i)[0][1] == dict2.get(i)[0][1]:
                    self.dict[i] = dict2.get(i)
            return self.dict



import unittest
from ...files.DefaultDictTTL_class import DefaultDictTTL
import time
from unittest.mock import patch
import math
r
class TestDefaultDictTTL(unittest.TestCase):
    def setUp(self):
        self.data = [('a', 10), ('a', 20), ('blue', 2), ('blue', 4)]
        self.dict_ttl = DefaultDictTTL(10)
        [self.dict_ttl.append_values(k, v) for k, v in self.data]

        self.new_data = [('c', 3), ('c', 5),('a',10),('a',20)]
        self.dict_ttl1 = DefaultDictTTL(3)
        [self.dict_ttl1.append_values(k, v) for k, v in self.new_data]

    @patch('time.sleep', return_value=None)
    def test_is_expired(self, patched_time_sleep):
        now = time.time()
        self.assertFalse(self.dict_ttl.is_expired('a'))
        self.assertFalse(self.dict_ttl.is_expired('a',now=now))
        self.assertTrue(len(self.dict_ttl),2)
        time.sleep(10)
        self.assertTrue(self.dict_ttl.is_expired('a',now=now + 11))
        self.assertFalse(self.dict_ttl.is_expired('a', now=now + 9))
        self.assertTrue(len(self.dict_ttl), 1)

    @patch('time.sleep', return_value=None)
    def test__purge(self, patched_time_sleep):
        self.assertTrue(len(self.dict_ttl),2)
        time.sleep(11)
        self.dict_ttl._purge()
        self.assertTrue(len(self.dict_ttl),0)

    def test_get(self):
        self.assertEqual(self.dict_ttl.get('a')[0][1],[10,20])
        self.assertEqual(self.dict_ttl.get('blue')[0][1],[2,4])

        self.assertNotEqual(self.dict_ttl.get('a')[0][1], 10)
        self.assertNotEqual(self.dict_ttl.get('blue')[0][1], 2)

    def test_set_ttl(self):
        now = time.time()
        self.dict_ttl.set_ttl('a',10)
        self.assertEqual(math.floor(self.dict_ttl.get('a')[0][0]),math.floor(now + 10))
        self.assertNotEqual(math.floor(self.dict_ttl.get('a')[0][0]),math.floor(now + 11))

    def test_get_ttl(self):
        self.assertEqual(round(self.dict_ttl.get_ttl('a'),0),10)
        self.assertNotEqual(round(self.dict_ttl.get_ttl('a'),0),11)

    def test_expire_at(self):
        now = time.time()
        self.dict_ttl.expire_at('a',now + 10)
        self.assertEqual(math.ceil(self.dict_ttl.get_ttl('a')),10)
        self.assertNotEqual(math.ceil(self.dict_ttl.get_ttl('a')),11)

    def test__iter__(self):
        self.data = [('a',[10,20]),('blue',[2,4])]
        self.false_data = [('a',10),('b',30)]
        self.dict_data = list(self.dict_ttl.items())
        self.assertEqual(self.data,self.dict_data)
        self.assertNotEqual(self.data,self.false_data)

    def test__set_item(self):
        self.data_items = [('c',[10]),('d',[20])]
        self.false_data_items = [('c',30),('d',40)]
        self.dict_new_ttl = DefaultDictTTL(10)
        self.dict_new_ttl['c'] = 10
        self.dict_new_ttl['d'] = 20
        self.dict_data = list(self.dict_new_ttl.items())
        self.assertEqual(self.data_items, self.dict_data)
        self.assertNotEqual(self.false_data_items, self.dict_data)

    def test__delitem__(self):
        self.assertIn('a', self.dict_ttl.keys())
        self.dict_ttl.__delitem__('a')
        self.assertNotIn('a',self.dict_ttl.keys())

    def test_delete_key_item(self):
        self.assertIn(10,self.dict_ttl.get('a')[0][1])
        self.dict_ttl.delete_key_item('a',10)
        self.assertNotIn(10, self.dict_ttl.get('a')[0][1])

    def test__len__(self):
        self.assertEqual(len(self.dict_ttl),2)
        self.assertNotEqual(len(self.dict_ttl),3)

    def test__getitem__(self):
        self.assertEqual(self.dict_ttl.__getitem__('a')[0][1],[10,20])
        self.assertNotEqual(self.dict_ttl.__getitem__('b'),10)

    def test_keys(self):
        self.assertEqual(list(self.dict_ttl.keys()),['a','blue'])
        self.assertNotEqual(list(self.dict_ttl.keys()), ['a', 'c'])

    def test_items(self):
        self.data = [('a',[10,20]),('blue',[2,4])]
        self.false_data = [('a',10),('b',20)]
        self.assertEqual(self.dict_ttl.items(),self.data)
        self.assertNotEqual(self.dict_ttl.items(), self.false_data)

    def test_ttl_items(self):
        now = time.time()
        self.data = [('a',[(math.ceil(now) + 10,[10,20])]),('blue',[(math.ceil(now) + 10,[2,4])])]
        self.false_data = [('c',(math.ceil(now) + 20,10)),('d',(math.ceil(now) + 20,20))]

        [self.assertEqual(i[0],j[0]) for i,j in zip(self.dict_ttl.ttl_items(),self.data)]
        [self.assertEqual(math.ceil(i[1][0][0]),math.ceil(j[1][0][0])) for i, j in zip(self.dict_ttl.ttl_items(), self.data)]
        [self.assertEqual(i[1][0][1], j[1][0][1]) for i, j in zip(self.dict_ttl.ttl_items(), self.data)]

    def test_values(self):
        self.data = [[10,20],[2,4]]
        self.false_data = [3,4]
        for k,v in zip(self.dict_ttl.values(),self.data):
            self.assertEqual(k[0][1],v)

        for k,v in zip(self.dict_ttl.values(),self.false_data):
            self.assertNotEqual(k[0][1],v)

    def test_values_without_ttl(self):
        self.assertEqual(self.dict_ttl.values_without_ttl(),[[10,20],[2,4]])
        self.assertNotEqual(self.dict_ttl.values_without_ttl(),[1,2,3])

    def test_dict_union(self):
        self.dict_ttl2 = DefaultDictTTL(3).dict_union(self.dict_ttl,self.dict_ttl1)
        self.assertEqual(list(self.dict_ttl2.keys()),['a','blue','c'])
        self.values_new = [i[0][1] for i in list(self.dict_ttl2.values())]
        self.assertEqual(self.values_new,[[10,20],[2,4],[3,5]])

    def test_dict_intersection(self):
        self.dict_ttl2 = DefaultDictTTL(3).dict_intersection(self.dict_ttl, self.dict_ttl1)
        self.assertEqual(list(self.dict_ttl2.keys()), ['a'])
        self.values_new = [i[0][1] for i in list(self.dict_ttl2.values())]
        self.assertEqual(self.values_new, [[10, 20]])


if __name__ == '__main__':
    unittest.TestCase()
